# SPDX-FileCopyrightText: 2024 DB Systel GmbH
#
# SPDX-License-Identifier: Apache-2.0

"""Create a CycloneDX SBOM using cgxgen as Docker container"""

import logging
import re
import sys
from os.path import abspath, basename, dirname
from shutil import copy2
from tempfile import NamedTemporaryFile, gettempdir
from uuid import uuid4

import docker
from docker.errors import APIError, ContainerError, DockerException, ImageNotFound


def _sanitize_container_name(name: str) -> str:
    """
    Sanitizes a Docker container name to conform to Docker's naming rules.

    Args:
        name (str): The initial container name, potentially containing invalid
        characters.

    Returns:
        str: The sanitized container name, with spaces replaced and disallowed
        characters removed.
    """
    # Remove spaces
    name = name.replace(" ", "_")
    # Remove all disallowed characters
    name = re.sub("[^a-zA-Z0-9_.-]+", "_", name)
    # Fix first character when it's not allowed, and return
    return re.sub("^[^a-zA-Z0-9]+", "0", name)


def _run_cdxgen(
    dclient: docker.DockerClient,
    directory: str,
    cont_name: str,
    output_path: str,
    image: str = "ghcr.io/cyclonedx/cdxgen",
) -> None:
    """
    Runs a Docker container using the specified cdxgen image to generate a SBOM
    for the provided directory.

    Args:
        dclient (docker.DockerClient): The Docker client used to interact with
        the Docker service.

        directory (str): The path to the project directory for which the SBOM is
        to be generated.

        cont_name (str): The name assigned to the Docker container.

        output_path (str): The file path where the generated SBOM will be saved
        on the local system.

        image (str, optional): The Docker image to use for generating the SBOM.
        Defaults to `ghcr.io/cyclonedx/cdxgen`.

    Raises:
        SystemExit: If the Docker container fails to start, the image is not
        found, or there is a Docker API error.
    """
    logging.debug(
        "Running image '%s' on directory '%s' with name '%s' and output '%s'",
        image,
        directory,
        cont_name,
        output_path,
    )
    try:
        dclient.containers.run(
            image=image,
            name=cont_name,
            remove=True,
            volumes=[f"{directory}:/app", f"{dirname(output_path)}:/sbom_data"],
            tty=True,
            command=["-r", "/app", "-o", f"/sbom_data/{basename(output_path)}"],
        )
    except ContainerError as err:
        logging.critical("Docker container wasn't able to start: %s", err)
        sys.exit(1)
    except ImageNotFound as err:
        logging.critical("Docker image not found: %s", err)
        sys.exit(1)
    except APIError as err:
        logging.critical("Docker API error: %s", err)
        sys.exit(1)


def generate_cdx_sbom(directory: str, output: str = "") -> str:
    """
    Generates a CycloneDX Software Bill of Materials (SBOM) for the project
    located in the specified directory.

    This function uses the Docker `cdxgen` image, a tool for generating SBOMs,
    inside a container. The resulting SBOM is saved as a JSON file and its path
    is returned.

    Args:
        directory (str): The path to the directory containing the project for
        which the SBOM is to be generated. The path can be either relative or
        absolute.

        output (str): The path to the SBOM that is to be generated. If left
        empty, it will be created in a temporary directory.

    Returns:
        str: The absolute path to the generated SBOM JSON file.

    Raises:
        SystemExit: If the Docker daemon is not available or cannot be reached,
        the function logs a critical error message and exits the program with
        status code 1.

    Dependencies:
        - Docker: The function interacts with Docker to create and manage
          containers.
    """
    # Initiate Docker client
    try:
        dclient = docker.from_env()
    except DockerException as err_daemon:
        logging.critical("Docker daemon does not seem to be available: %s", err_daemon)
        sys.exit(1)

    # Turn directory into absolute path, to support directories like `.`
    directory = str(abspath(directory))
    # Define names for container and SBOM output
    cont_name = _sanitize_container_name(f"{basename(directory)}_{uuid4().hex[:6]}")
    # Define output path
    if not output:
        output = f"{gettempdir()}/{cont_name}.json"

    # Create SBOM in temporary file. This enables the program to delete the SBOM
    # generated by cdxgen under root ownership
    logging.info("Generating SBOM for %s using cdxgen", directory)
    with NamedTemporaryFile() as tmpfile:
        _run_cdxgen(dclient, directory, cont_name, tmpfile.name)

        # Copy to final destination with user permissions
        copy2(tmpfile.name, output)

        logging.info("SBOM has been saved to %s", output)

    return output
